import sys
import traceback
from contextlib import contextmanager
from contextvars import ContextVar
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Optional

from ajprax.experimental.subscriptions import Events
from ajprax.print import print

TRACE = 0
DEBUG = 10
INFO = 20
WARN = 30
ERROR = 40
FATAL = 50

LEVEL_NAME = {
    TRACE: "TRACE",
    DEBUG: "DEBUG",
    INFO: " INFO",
    WARN: " WARN",
    ERROR: "ERROR",
    FATAL: "FATAL",
}


@dataclass
class Log:
    datetime: datetime
    level: int
    message: str
    keywords: dict
    exception: Optional[BaseException]

    def __post_init__(self):
        if isinstance(self.exception, bool):
            if self.exception:
                self.exception = sys.exc_info()[1]
            else:
                self.exception = None

    def __str__(self):
        message = str(self.message)
        ts = self.datetime.isoformat().replace("+00:00", "Z")
        level = LEVEL_NAME[self.level]
        if self.keywords:
            message += " " if message else ""
            message += " ".join(f"{k}={repr(v)}" for k, v in self.keywords.items())
        if self.exception:
            message += "\n"
            message += "".join("\t" + line for line in traceback.format_exception(self.exception))
        return f"{ts} {level} {message}"


class Logger(Events):
    def __init__(self):
        Events.__init__(self)
        self._global_level = INFO
        self._context_level = ContextVar("level", default=None)
        self._kwargs = ContextVar("kwargs", default={})
        self.subscribe(print)

    @property
    def level(self):
        context_level = self._context_level.get()
        if context_level is None:
            return self._global_level
        return context_level

    @level.setter
    def level(self, level):
        self._global_level = level

    def _log(self, _level, _message, _exception, **kwargs):
        if _level >= self.level:
            self.send(Log(
                datetime.now(timezone.utc),
                _level,
                _message,
                {**self._kwargs.get(), **kwargs},
                _exception,
            ))

    def trace(self, _message="", _exception=False, **kwargs):
        self._log(TRACE, _message, _exception, **kwargs)

    def debug(self, _message="", _exception=False, **kwargs):
        self._log(DEBUG, _message, _exception, **kwargs)

    def info(self, _message="", _exception=False, **kwargs):
        self._log(INFO, _message, _exception, **kwargs)

    def warn(self, _message="", _exception=False, **kwargs):
        self._log(WARN, _message, _exception, **kwargs)

    def error(self, _message="", _exception=False, **kwargs):
        self._log(ERROR, _message, _exception, **kwargs)

    def fatal(self, _message="", _exception=False, **kwargs):
        self._log(FATAL, _message, _exception, **kwargs)

    @contextmanager
    def context_level(self, level):
        token = self._context_level.set(level)
        try:
            yield
        finally:
            self._context_level.reset(token)

    @contextmanager
    def context_kwargs(self, **kw):
        token = self._kwargs.set(kw)
        try:
            yield
        finally:
            self._kwargs.reset(token)


log = Logger()



import time
from functools import partial
from threading import Event, Thread, Condition

from ajprax.cache import cache, InProgress, Cache
from ajprax.logging import log
from ajprax.sentinel import Unset
from tests import should_raise


def add(a, b):
    return a + b


class TestCache:
    def _test_default_key(self, f):
        def test(expected_count, expected_value, f, *a, **kw):
            actual_value = f(*a, **kw)
            actual_count = self.count
            assert actual_value == expected_value
            assert actual_count == expected_count

        self.count = 0
        test(1, [1, 1], f, 1, 2)
        test(1, [1, 1], f, 1, b=2)
        test(1, [1, 1], f, a=1, b=2)
        test(2, [1, 1, 1], f, 1, 3)
        test(3, [2], f, 2, 1)

    def _test_custom_key(self, f):
        def test(expected_count, expected_value, f, *a, **kw):
            actual_value = f(*a, **kw)
            actual_count = self.count
            assert actual_value == expected_value
            assert actual_count == expected_count

        self.count = 0
        test(1, [1, 1], f, 1, 2)
        test(1, [1, 1], f, 1, b=2)
        test(1, [1, 1], f, a=1, b=2)
        test(2, [1, 1, 1], f, 1, 3)
        test(2, [1, 1], f, 2, 1)

    def test_function(self):
        @cache()
        def f(a, b):
            self.count += 1
            return [a] * b

        self._test_default_key(f)

    def test_function_no_call(self):
        @cache
        def f(a, b):
            self.count += 1
            return [a] * b

        self._test_default_key(f)

    def test_function_custom(self):
        @cache(key=add)
        def f(a, b):
            self.count += 1
            return [a] * b

        self._test_custom_key(f)

    def test_instance_method(self):
        class C:
            @cache(method=True)
            def f(_, a, b):
                self.count += 1
                return [a] * b

        self._test_default_key(C().f)

    def test_instance_method_custom(self):
        class C:
            @cache(key=add, method=True)
            def f(_, a, b):
                self.count += 1
                return [a] * b

        self._test_custom_key(C().f)

    def test_instance_method_partial(self):
        class C:
            @cache(method=True)
            def f(_, a, b):
                self.count += 1
                return [a] * b

        self._test_default_key(partial(C.f, C()))

    def test_instance_method_partial_custom(self):
        class C:
            @cache(key=add, method=True)
            def f(_, a, b):
                self.count += 1
                return [a] * b

        self._test_custom_key(partial(C.f, C()))

    def test_classmethod(self):
        class C:
            @classmethod
            @cache(method=True)
            def f(_, a, b):
                self.count += 1
                return [a] * b

        self._test_default_key(C.f)

    def test_classmethod_custom(self):
        class C:
            @classmethod
            @cache(key=add, method=True)
            def f(_, a, b):
                self.count += 1
                return [a] * b

        self._test_custom_key(C.f)

    def test_classmethod_via_instance(self):
        class C:
            @classmethod
            @cache(method=True)
            def f(_, a, b):
                self.count += 1
                return [a] * b

        self._test_default_key(C().f)

    def test_classmethod_via_instance_custom(self):
        class C:
            @classmethod
            @cache(key=add, method=True)
            def f(_, a, b):
                self.count += 1
                return [a] * b

        self._test_custom_key(C().f)

    def test_classmethod_subclass(self):
        class C:
            @classmethod
            @cache(method=True)
            def f(_, a, b):
                self.count += 1
                return [a] * b

        class D(C):
            pass

        self._test_default_key(D.f)

        # C.f should have nothing in it, so calling it will increase self.count
        count = self.count
        C.f(1, 2)
        assert self.count == count + 1

    def test_classmethod_subclass_custom(self):
        class C:
            @classmethod
            @cache(key=add, method=True)
            def f(_, a, b):
                self.count += 1
                return [a] * b

        class D(C):
            pass

        self._test_custom_key(D.f)

        # C.f should have nothing in it, so calling it will increase self.count
        count = self.count
        C.f(1, 2)
        assert self.count == count + 1

    def test_classmethod_subclass_via_instance(self):
        class C:
            @classmethod
            @cache(method=True)
            def f(_, a, b):
                self.count += 1
                return [a] * b

        class D(C):
            pass

        self._test_default_key(D().f)

        # the cache should live with the class
        count = self.count
        D.f(1, 2)
        assert self.count == count

        # C.f should have nothing in it, so calling it will increase self.count
        count = self.count
        C.f(1, 2)
        assert self.count == count + 1

    def test_classmethod_subclass_via_instance_custom(self):
        class C:
            @classmethod
            @cache(key=add, method=True)
            def f(_, a, b):
                self.count += 1
                return [a] * b

        class D(C):
            pass

        self._test_custom_key(D().f)

        # the cache should live with the class
        count = self.count
        D.f(1, 2)
        assert self.count == count

        # C.f should have nothing in it, so calling it will increase self.count
        count = self.count
        C.f(1, 2)
        assert self.count == count + 1

    def test_staticmethod(self):
        class C:
            @staticmethod
            @cache()
            def f(a, b):
                self.count += 1
                return [a] * b

        self._test_default_key(C.f)

    def test_staticmethod_no_call(self):
        class C:
            @staticmethod
            @cache
            def f(a, b):
                self.count += 1
                return [a] * b

        self._test_default_key(C.f)

    def test_staticmethod_custom(self):
        class C:
            @staticmethod
            @cache(key=add)
            def f(a, b):
                self.count += 1
                return [a] * b

        self._test_custom_key(C.f)

    def test_staticmethod_via_instance(self):
        class C:
            @staticmethod
            @cache()
            def f(a, b):
                self.count += 1
                return [a] * b

        self._test_default_key(C().f)

        # the cache should live with the class
        count = self.count
        C.f(1, 2)
        assert self.count == count

    def test_staticmethod_via_instance_no_call(self):
        class C:
            @staticmethod
            @cache
            def f(a, b):
                self.count += 1
                return [a] * b

        self._test_default_key(C().f)

        # the cache should live with the class
        count = self.count
        C.f(1, 2)
        assert self.count == count

    def test_staticmethod_via_instance_custom(self):
        class C:
            @staticmethod
            @cache(key=add)
            def f(a, b):
                self.count += 1
                return [a] * b

        self._test_custom_key(C().f)

        # the cache should live with the class
        count = self.count
        C.f(1, 2)
        assert self.count == count

    def test_staticmethod_subclass(self):
        class C:
            @staticmethod
            @cache()
            def f(a, b):
                self.count += 1
                return [a] * b

        class D(C):
            pass

        self._test_default_key(D.f)

        # for static methods, the cache is shared with subclasses
        count = self.count
        C.f(1, 2)
        assert self.count == count

    def test_staticmethod_subclass_no_call(self):
        class C:
            @staticmethod
            @cache
            def f(a, b):
                self.count += 1
                return [a] * b

        class D(C):
            pass

        self._test_default_key(D.f)

        # for static methods, the cache is shared with subclasses
        count = self.count
        C.f(1, 2)
        assert self.count == count

    def test_staticmethod_subclass_custom(self):
        class C:
            @staticmethod
            @cache(key=add)
            def f(a, b):
                self.count += 1
                return [a] * b

        class D(C):
            pass

        self._test_custom_key(D.f)

        # for static methods, the cache is shared with subclasses
        count = self.count
        C.f(1, 2)
        assert self.count == count

    def test_staticmethod_subclass_via_instance(self):
        class C:
            @staticmethod
            @cache()
            def f(a, b):
                self.count += 1
                return [a] * b

        class D(C):
            pass

        self._test_default_key(D().f)

        # for static methods, the cache is shared with subclasses
        count = self.count
        C.f(1, 2)
        assert self.count == count

    def test_staticmethod_subclass_via_instance_no_call(self):
        class C:
            @staticmethod
            @cache
            def f(a, b):
                self.count += 1
                return [a] * b

        class D(C):
            pass

        self._test_default_key(D().f)

        # for static methods, the cache is shared with subclasses
        count = self.count
        C.f(1, 2)
        assert self.count == count

    def test_staticmethod_subclass_via_instance_custom(self):
        class C:
            @staticmethod
            @cache(key=add)
            def f(a, b):
                self.count += 1
                return [a] * b

        class D(C):
            pass

        self._test_custom_key(D().f)

        # for static methods, the cache is shared with subclasses
        count = self.count
        C.f(1, 2)
        assert self.count == count

    def test_default_values(self):
        @cache()
        def f(a, b=2):
            self.count += 1
            return [a] * b

        def test(expected_count, expected_value, f, *a, **kw):
            actual_value = f(*a, **kw)
            actual_count = self.count
            assert actual_value == expected_value
            assert actual_count == expected_count

        self.count = 0
        test(1, [1, 1], f, 1)
        test(1, [1, 1], f, 1)
        test(1, [1, 1], f, a=1)
        test(2, [1, 1, 1], f, 1, 3)
        test(3, [2], f, 2, 1)

    def test_default_values_no_call(self):
        @cache
        def f(a, b=2):
            self.count += 1
            return [a] * b

        def test(expected_count, expected_value, f, *a, **kw):
            actual_value = f(*a, **kw)
            actual_count = self.count
            assert actual_value == expected_value
            assert actual_count == expected_count

        self.count = 0
        test(1, [1, 1], f, 1)
        test(1, [1, 1], f, 1)
        test(1, [1, 1], f, a=1)
        test(2, [1, 1, 1], f, 1, 3)
        test(3, [2], f, 2, 1)

    def test_default_values_custom(self):
        @cache(key=add)
        def f(a, b=2):
            self.count += 1
            return [a] * b

        def test(expected_count, expected_value, f, *a, **kw):
            actual_value = f(*a, **kw)
            actual_count = self.count
            assert actual_value == expected_value
            assert actual_count == expected_count

        self.count = 0
        test(1, [1, 1], f, 1)
        test(1, [1, 1], f, 1)
        test(1, [1, 1], f, a=1)
        test(2, [1, 1, 1], f, 1, 3)
        test(2, [1, 1], f, 2, 1)

    def test_two_instances(self):
        class C:
            @cache(method=True)
            def f(_, a, b):
                self.count += 1
                return [a] * b

        c = C()
        c2 = C()

        self._test_default_key(c.f)
        self._test_default_key(c2.f)

    def test_raise_for_all_callers(self):
        leader_has_started = Event()
        follower_is_waiting = Event()

        def f(_):
            leader_has_started.set()
            follower_is_waiting.wait()
            raise ValueError

        f = Cache(f, Unset, False, follower_is_waiting)

        def target():
            with should_raise(ValueError):
                f(None)

        leader = Thread(target=target)
        follower = Thread(target=target)
        leader.start()
        leader_has_started.wait()
        follower.start()
        leader.join()
        follower.join()

    def test_succeeds_after_failing(self):
        @cache
        def f(_):
            if should_fail:
                raise ValueError
            return "success"

        should_fail = True
        with should_raise(ValueError):
            f(None)
        should_fail = False
        assert f(None) == "success"


class TestSingleton:

    def _test(self, f):
        self.count = 0
        f()
        assert self.count == 1
        f()
        assert self.count == 1

    def test_function(self):
        @cache()
        def f():
            self.count += 1

        self._test(f)

    def test_instance_method(self):
        class C:
            @cache(method=True)
            def f(_):
                self.count += 1

        c = C()
        self._test(c.f)

    def test_instance_method_partial(self):
        class C:
            @cache(method=True)
            def f(_):
                self.count += 1

        c = C()
        self._test(partial(C.f, c))

    def test_classmethod(self):
        class C:
            @classmethod
            @cache(method=True)
            def f(cls):
                self.count += 1

        self._test(C.f)

    def test_classmethod_via_instance(self):
        class C:
            @classmethod
            @cache(method=True)
            def f(cls):
                self.count += 1

        c = C()
        self._test(c.f)

    def test_classmethod_subclass(self):
        class C:
            @classmethod
            @cache(method=True)
            def f(cls):
                self.count += 1

        class D(C):
            pass

        self._test(D.f)
        count = self.count
        C.f()
        assert self.count == count + 1

    def test_classmethod_subclass_via_instance(self):
        class C:
            @classmethod
            @cache(method=True)
            def f(cls):
                self.count += 1

        class D(C):
            pass

        self._test(D().f)
        count = self.count
        C.f()
        assert self.count == count + 1

    def test_staticmethod(self):
        class C:
            @staticmethod
            @cache()
            def f():
                self.count += 1

        self._test(C.f)

    def test_staticmethod_via_instance(self):
        class C:
            @staticmethod
            @cache()
            def f():
                self.count += 1

        self._test(C().f)

    def test_staticmethod_subclass(self):
        class C:
            @staticmethod
            @cache()
            def f():
                self.count += 1

        class D(C):
            pass

        self._test(D.f)

    def test_staticmethod_subclass_via_instance(self):
        class C:
            @staticmethod
            @cache()
            def f():
                self.count += 1

        class D(C):
            pass

        self._test(D().f)

    def test_two_instances(self):
        class C:
            @cache(method=True)
            def f(_):
                self.count += 1

        c = C()
        c2 = C()

        self._test(c.f)
        self._test(c2.f)

    def test_property(self):
        class C:
            @property
            @cache(method=True)
            def f(_):
                self.count += 1

        self.count = 0
        c = C()
        c.f
        assert self.count == 1
        c.f
        assert self.count == 1



from datetime import datetime, timezone
from threading import Thread, Event

from ajprax.collections import Iter
from ajprax.logging import LEVEL_NAME, Logger, Log, INFO, WARN


class TestLog:
    def test_level(self):
        def test(log_level, message_level, *a, **kw):
            def check(log):
                # log level is passed through to the Log object
                assert log.level == message_level
                # log is only sent if the message level meets or exceeds the log level
                assert message_level >= log_level

            logger = Logger()
            logger.level = log_level
            logger.subscribe(check)
            name = LEVEL_NAME[message_level].strip().lower()
            getattr(logger, name)(*a, **kw)

        for log_level, message_level in Iter(LEVEL_NAME).product(repeat=2):
            test(log_level, message_level)
            test(log_level, message_level, "message")
            test(log_level, message_level, key="value")
            test(log_level, message_level, "message", key="value")

    def test_format(self):
        def test(datetime, level, message, keywords, exception):
            log = str(Log(datetime, level, message, keywords, exception))
            assert ("log message" in log) == bool(message)
            assert ("key='value'" in log) == bool(keywords)
            assert ("exception message" in log) == bool(exception)
            assert log.split("\n")[0].startswith("2024-01-01T00:00:00.000001Z  INFO ")

        for (message, keywords, exception) in Iter((True, False)).product(repeat=3):
            message = "log message" if message else ""
            keywords = {"key": "value"} if keywords else {}
            exception = Exception("exception message") if exception else None
            test(
                datetime(2024, 1, 1, 0, 0, 0, 1, tzinfo=timezone.utc),
                INFO,
                message,
                keywords,
                exception,
            )

    def test_context_level_prevents(self):
        def a():
            with log.context_level(WARN):
                bevent.set()
                log.info("a")
                aevent.wait()

        def b():
            bevent.wait()
            log.info("b")
            aevent.set()

        def test(l):
            assert l.message == "b"

        log = Logger()
        log.subscribe(test)

        aevent = Event()
        bevent = Event()
        athread = Thread(target=a)
        bthread = Thread(target=b)
        athread.start()
        bthread.start()
        athread.join()
        bthread.join()

    def test_context_level_allows(self):
        def a():
            with log.context_level(INFO):
                bevent.set()
                log.info("a")
                aevent.wait()

        def b():
            bevent.wait()
            log.info("b")
            aevent.set()

        def test(l):
            assert l.message == "a"

        log = Logger()
        log.subscribe(test)
        log.level = WARN

        aevent = Event()
        bevent = Event()
        athread = Thread(target=a)
        bthread = Thread(target=b)
        athread.start()
        bthread.start()
        athread.join()
        bthread.join()

    def test_context_kwargs(self):
        def a():
            with log.context_kwargs(a=True):
                bevent.set()
                aevent.wait()

        def b():
            bevent.wait()
            log.info(b=True)
            aevent.set()

        def test(l):
            assert "a" not in l.keywords
            assert "b" in l.keywords

        log = Logger()
        log.subscribe(test)

        aevent = Event()
        bevent = Event()
        athread = Thread(target=a)
        bthread = Thread(target=b)
        athread.start()
        bthread.start()
        athread.join()
        bthread.join()



from functools import partial
from inspect import signature
from threading import Lock, Condition

from ajprax.require import require
from ajprax.sentinel import Unset


def cache(f=None, *, key=Unset, method=False):
    """
    Major differences from functools.cache:
    - Cached function will not be called more than once for the same key, even if a second call happens before the
      first completes. Different keys can still be called concurrently.
    - Allows customizing key generation.

    Works with bare functions, instance methods, classmethods, and staticmethods.
    Instance methods and classmethods should use method=True. This will store the cache on the instance for isolation
      and so that the cache can be garbage collected with the instance. This also excludes the instance or class from
      the key function so that methods on unhashable types can still be cached.

    Handles arguments uniformly, including default values and arguments specified as any mix of positional and keyword.

    Default key function produces a tuple of arguments, so all arguments must be hashable
      (except self/cls if method=True).

    Automatically optimizes implementation when used to create a singleton.
    """
    if f is not None:
        return cache(key=key, method=method)(f)

    def decorator(f):
        if len(signature(f).parameters) == method:
            require(key is Unset, "cannot use custom key function for function with no arguments")
            return Singleton(f, method)
        return Cache(f, key, method)

    return decorator


class Cell:
    """Container which can be attached to an instance or class for cache isolation"""

    def __init__(self, value=Unset):
        self.value = value
        self.lock = Lock()


class Singleton:
    def __init__(self, f, method):
        self.f = f
        self.method = method
        self.creation_lock = Lock()
        self.name = f"_{f.__name__}_singleton"

    def __get__(self, instance, owner=None):
        return self if instance is None else partial(self, instance)

    def get_or_create_cell(self, obj):
        def get(obj):
            return getattr(obj, self.name, None)

        if not get(obj):
            with self.creation_lock:
                if not get(obj):
                    setattr(obj, self.name, Cell())
        return get(obj)

    def __call__(self, *a, **kw):
        cell = self.get_or_create_cell(a[0] if self.method else self)
        if cell.value is Unset:
            with cell.lock:
                if cell.value is Unset:
                    cell.value = self.f(*a, **kw)
        return cell.value


class DefaultKey:
    def __init__(self, parameters):
        self.parameters = parameters

    def __call__(self, *a, **kw):
        return *a, *(kw[param] for param in self.parameters if param in kw)


class InProgress(Condition):
    """
    Inserted into a cache before starting to generate the value so that concurrent callers can wait for the value
    instead of redundantly calling the cached function.

    Wrapped so that we don't mistake user Condition values for our marker
    """

    def __init__(self, lock):
        Condition.__init__(self, lock)
        self.exception = None


class Cache:
    def __init__(self, f, key, method, waiting_for_in_progress=None):
        self.f = f
        self.signature = signature(f)
        self.key = DefaultKey(tuple(self.signature.parameters)) if key is Unset else key
        self.method = method
        # Event used for synchronization in tests
        self.waiting_for_in_progress = waiting_for_in_progress
        self.name = f"_{f.__name__}_cache"
        self.creation_lock = Lock()

    def get_or_create_cell(self, obj):
        def get(obj):
            return getattr(obj, self.name, None)

        if not get(obj):
            with self.creation_lock:
                if not get(obj):
                    setattr(obj, self.name, Cell({}))
        return get(obj)

    def __get__(self, instance, owner=None):
        return self if instance is None else partial(self, instance)

    def __call__(self, *a, **kw):
        args = self.signature.bind(*a, **kw)
        args.apply_defaults()
        key = self.key(*args.args[self.method:], **args.kwargs)
        cell = self.get_or_create_cell(a[0] if self.method else self)

        with cell.lock:
            existing = cell.value.get(key, Unset)
            if existing is Unset:
                condition = InProgress(cell.lock)
                cell.value[key] = condition
            elif isinstance(existing, InProgress):
                condition = existing
                while condition.exception is None and cell.value.get(key) is condition:
                    if self.waiting_for_in_progress:
                        self.waiting_for_in_progress.set()
                    condition.wait()
                # raise exception in all waiting threads
                if condition.exception is not None:
                    raise condition.exception
                return cell.value[key]
            else:
                return existing

        try:
            value = self.f(*args.args, **args.kwargs)
        except Exception as e:
            with cell.lock:
                if cell.value.get(key) is condition:
                    condition.exception = e
                    condition.notify_all()
                    del cell.value[key]
            raise

        with cell.lock:
            # it's important to set this value under the lock to avoid a race condition where a concurrent caller
            # gets the InProgress from the dict before we overwrite it, but waits on it after we notify_all.
            # in short, all reads and writes to the dict should be under the lock.
            cell.value[key] = value
            condition.notify_all()
        return value
